package EmailSystem; 

import java.util.ArrayList; 
import java.util.Collections; 
import java.util.List; 

import TestSpecifications.SpecificationManager; 

public   class  Client {
	
	protected int id;

	
	protected String name;

	 // name does also serve as address at the moment

	public int getId() {
		return id;
	}

	
	
	// incoming emails reach the user at this point. here they are put in a
	// mailbox.
	static void deliver(Client client, Email msg) {
		Util.prompt("mail delivered\n");
	}

	

	// incoming emails are processed by this method before delivery.
	private static void  incoming__wrappee__encrypt  (Client client, Email msg) {
		deliver(client, msg);
	}

	

	private static void  incoming__wrappee__forward  (Client client, Email msg) {
		incoming__wrappee__encrypt(client, msg);
		Client receiver = client.getForwardReceiver();
		if (receiver != null) {
			msg.setEmailTo(receiver.getName());
			forward(client, msg);
			incoming(receiver, msg);
		}
	}

	
	static void incoming(Client client, Email msg) {
		// decrypt

		int privkey = client.getPrivateKey();
		if (privkey != 0) {
			if (msg.isEncrypted()
					&& isKeyPairValid(msg.getEmailEncryptionKey(), privkey)) {
				msg.setEmailIsEncrypted(false);
				msg.setEmailEncryptionKey(0);
			}
		}
		// end decrypt

		incoming__wrappee__forward(client, msg);
	}

	

	// outgoing emails leave the client at this point. here they are put in an
	// outgoing queue instead.
	static void mail(Client client, Email msg) {
		Util.prompt("mail sent");
	}

	


	// emails to be sent are processed by this method before beeing mailed.
	private static void  outgoing__wrappee__keys  (Client client, Email msg) {
		msg.setEmailFrom(client);
		mail(client, msg);
	}

	
	// emails to be sent are processed by this method before being mailed.
	static void outgoing(Client client, Email msg) {

		// encrypt
		Client receiver = getClientByAdress(msg.getEmailTo());
		int pubkey = client.getKeyringPublicKeyByClient(receiver);
		if (pubkey != 0) {
			msg.setEmailEncryptionKey(pubkey);
			msg.setEmailIsEncrypted(true);
			Util.prompt("Encrypted Mail " + msg.getId());
		}
		// msg.setEmailIsEncrypted(true); // von mir gelöscht, das macht keinen
		// sinn!

		// end encrypt

		outgoing__wrappee__keys(client, msg);
	}

	

	public static int sendEmail(Client sender, String receiverAddress,
			String subject, String body) {
		Email email = Email.createEmail(sender, receiverAddress, subject, body);
		Util.prompt("sending Mail " + email.getId());
		outgoing(sender, email);
		Client receiver = Client.getClientByAdress(email.getEmailTo());
		if (receiver != null) {
			incoming(receiver, email);
		} else {
			throw new IllegalArgumentException("Receiver " + receiverAddress + " Unknown");
		}
		return 0; // die Zeile kommt von mir
	}

	

	private Client(int id, String name) {
		this.id = id;
		this.name = name;
	}

	

	public String getName() {
		return name;
	}

	

	static int clientCounter = 1;

	

	public static Client createClient(String name) {
		Client client = new Client(clientCounter++, name);
		clients[client.getId()] = client;
		return client;
	}

	

	static Client[] clients = new Client[4];

	

	static Client getClientById(int id) {
		return clients[id];
	}

	

	/**
	 * address equals name for now
	 * 
	 * This method implements the EmailAddress to Receiver lookup. The method
	 * returns null if the address is not found!
	 */
	static Client getClientByAdress(String address) {
		for (int i = 0; i < clients.length; i++) {
			if (clients[i] != null && clients[i].getName().equals(address)) {
				return clients[i];
			}
		}
		return null;
	}

	

	public static void resetClients() {
		clientCounter = 1;
		for (int i = 0; i < clients.length; i++) {
			clients[i] = null;
		}
	}

	

	@Override
	public String toString() {
		return name;
	}

	
	/**
	 * Saves public keys of other clients
	 */
	protected ArrayList<KeyringEntry> keyring = new ArrayList<KeyringEntry>();

	

	protected int privateKey;

	

	public void setPrivateKey(int privateKey) {
		this.privateKey = privateKey;
	}

	

	public int getPrivateKey() {
		return privateKey;
	}

	

	/**
	 * Currently this method does only set the private key to seed
	 */
	public static void generateKeyPair(Client client, int seed) {
		client.setPrivateKey(seed);
	}

	

	public void addKeyringEntry(Client client, int publicKey) {
		this.keyring.add(new KeyringEntry(client, publicKey));
	}

	
	
	/**
	 * Returns the public key of client, if it is saved in this keyring. If not
	 * the method returns 0.
	 */
	public int getKeyringPublicKeyByClient(Client client) {
		for (KeyringEntry e : keyring) {
			if (e.getKeyOwner().equals(client)) {
				return e.getPublicKey();
			}
		}
		return 0;
	}

	

	public static boolean isKeyPairValid(int publicKey, int privateKey) {
		Util.prompt("keypair valid " + publicKey + " " + privateKey);
		if (publicKey == 0 || privateKey == 0)
			return false;
		return privateKey == publicKey;
	}

	
	

	static  class  KeyringEntry {
		
		private Client keyOwner;

		
		private int publicKey;

		

		public KeyringEntry(Client keyOwner, int publicKey) {
			super();
			this.keyOwner = keyOwner;
			this.publicKey = publicKey;
		}

		

		public Client getKeyOwner() {
			return keyOwner;
		}

		

		public int getPublicKey() {
			return publicKey;
		}


	}

	
	protected Client forwardReceiver;

	
	
	public void setForwardReceiver(Client forwardReceiver) {
		this.forwardReceiver = forwardReceiver;
	}

	

	public Client getForwardReceiver() {
		return forwardReceiver;
	}

	

	static void forward(Client client, Email msg) {
		Util.prompt("Forwarding message.\n");
		Email.printMail(msg);
		outgoing(client, msg);
	}


}
