import gov.nasa.jpf.ListenerAdapter;

import gov.nasa.jpf.bdd.BddManager;

import gov.nasa.jpf.jvm.DynamicElementInfo;
import gov.nasa.jpf.jvm.JVM;
import gov.nasa.jpf.jvm.StaticElementInfo;
import gov.nasa.jpf.report.Statistics;
import gov.nasa.jpf.search.Search;

import net.sf.javabdd.BDD;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;

public class JPF_PrimitiveLogListener_Lifted_jpfBDD extends ListenerAdapter {
	private String resultMessagePrefix="jpfResultMessagePrefix: ";
	private String statesMessagePrefix="jpfStatesMessagePrefix: ";
	
	private int numSucc = 0;
	private int numErrors = 0;
	
	private BDD failedBdds = null;
	private BDD succeededBdds = null;
	
	private static final boolean logNumberOfAllErrors = true;
	private static final boolean limitToOneErrorPerProduct = true;
	
	// this line should be replaced by the Verification Environment
	//static String baseLogDir    = "/local/JPF_log/";
	private StaticElementInfo testRef;
	private String lastResult = null;
	private Search search;
	private int numSucceededBeforeError = 0;
		
	@Override
	public void searchFinished(Search search) {
		gov.nasa.jpf.Error jpfError = search.getLastError();
		//System.out.println(result==null?"result=null":"result="+result);
		if (jpfError != null && (lastResult == null || lastResult.startsWith("succeeded"))) {
			//System.out.println("Lifted prodcued error: " + jpfError.getDescription());
			DynamicElementInfo actions = (DynamicElementInfo) testRef.getFieldValueObject("actions");
			lastResult = "JPFerror: " + jpfError.getDescription() + 
			" Actions: " + actions.asString() + 
			"(" + (numSucc+numErrors) + " thread terminations during Verification)";
		}
		System.out.println("numErrors:" + numErrors + " numSucc:" + numSucc);
		if (failedBdds != null)
			System.out.println("Failed: " + BddManager.toAbbrevString(failedBdds));
		//(SatCount:" + failedBdds.allsat().size() + ")
		// satcount works, but it only counts the combinations of the present featureVars. Therefore it is misleading.
		if (succeededBdds != null)
			System.out.println("Succ: " + BddManager.toAbbrevString(succeededBdds));
		
		
		if (logNumberOfAllErrors) {
			// lastResult might be irrelevant (might be success even if there were errors before)
			if (numErrors == 0) {
				if (lastResult!=null)
					System.out.println(resultMessagePrefix + "succeeded (" + numSucc + " terminations)");
				else
					System.out.println(resultMessagePrefix + "No Thread Termination");
			} else {
				System.out.println(resultMessagePrefix + "Errors: " + numErrors + ", " + numSucc +" succeeded terminations");
			}
		} else {
			if (lastResult!=null)
				System.out.println(resultMessagePrefix + lastResult);
			else
				System.out.println(resultMessagePrefix + "No Thread Termination");
		}
		Statistics stat = search.getVM().getNextListenerOfType(Statistics.class, null);
		if (stat == null){
			System.out.println(statesMessagePrefix + "NA");
		} else {
			System.out.println(statesMessagePrefix + stat.newStates);
		}
	}

	@Override
	public void searchStarted(Search search) {
		System.out.println("searchClass is: " + search.getClass());
		this.search = search;
	}

	@Override
	public void threadTerminated(JVM vm) {
		// This means that one program execution path that was checked by JPF is terminated
		// one program execution path yields one TestConfiguration/LogEntry
		super.threadTerminated(vm);
		DynamicElementInfo errorInfo = (DynamicElementInfo) testRef.getFieldValueObject("error");
		DynamicElementInfo ignoreRun = (DynamicElementInfo) testRef.getFieldValueObject("ignoreThisRun");
		//DynamicElementInfo selectedFeatures = (DynamicElementInfo) testRef.getFieldValueObject("selectedFeatures");
		DynamicElementInfo actions = (DynamicElementInfo) testRef.getFieldValueObject("actions");
		
		gov.nasa.jpf.bdd.BddListener bddListener = vm.getJPF().getSearch().getNextListenerOfType(gov.nasa.jpf.bdd.BddListener.class, null);
		BDD bdd = bddListener.currentBDD;
		if (ignoreRun.asString().equals("true") || BddManager.isFalse(bdd)) // if the bdd is false, then this termination results from an "invalid" state that should not exist
			return;
		if (errorInfo == null || errorInfo.isNull()) {
			//System.out.println("Termination: Success");
			// succeeded
			numSucceededBeforeError++;
			lastResult = "succeeded (" + numSucceededBeforeError + ". Thread Termination)";
			if (this.succeededBdds != null)
				this.succeededBdds = this.succeededBdds.or(bdd);
			else this.succeededBdds = bdd;
			numSucc++;
		} else {
			//System.out.println("Termination: Failed: " + errorInfo.asString());
			//failed
			if (this.failedBdds != null)
				this.failedBdds = this.failedBdds.or(bdd);
			else this.failedBdds = bdd;
			String features = BddManager.toAbbrevString(bdd);
			lastResult = errorInfo.asString() + " Features: " + features + 
			" Actions: " + actions.asString() + 
			"(SuccededTerminationsBeforeThisError:"+numSucceededBeforeError+")";
			//System.out.println("Error: " + lastResult);
			numErrors++;
			
			if (!logNumberOfAllErrors) {
				// terminate after first error
				search.terminate();
			} else if (limitToOneErrorPerProduct) {
				//System.out.println("adding restriction: " + BddManager.toAbbrevString(bdd.not()));
				bddListener.addValidStatesRestriction(bdd.not());
			}
		}
	}
	
	@Override
	public void methodEntered(JVM vm) {
		super.methodEntered(vm);
		// this method is needed to get an initial reference to the test_Runner class
		// method search_started seems to be too early to do this (class is not yet initialized?)
		super.instructionExecuted(vm);
		if (testRef == null)
			testRef = vm.getClassReference("Test_Runner");
	}
}
